const PENDING  = 0,
      RESOLVED = 1,
      REJECTED = 2;

const REJECTED_REASON = Symbol('rejected_reason'),
      RESOLVED_VALUE = Symbol('resolved_value'),
      START_EXECUTOR = Symbol('start_executor'),
      ON_FULFILLED = Symbol('on_fulfilled'),
	  SUB_PLEDGES = Symbol('sub_pledges'),
      ON_REJECTED = Symbol('on_rejected'),
      DO_RESOLVE = Symbol('do_resolve'),
      DO_REJECT = Symbol('do_reject'),
      EXECUTOR = Symbol('executor'),
      UPDATED = Symbol('updated'),
      CREATED = Symbol('created'),
      STARTED = Symbol('started'),
      ENDED = Symbol('ended'),
      STATE = Symbol('state');

const SCHEDULE_EXECUTOR = Symbol('schedule_executor'),
      SCHEDULE_RESOLVE = Symbol('schedule_resolve'),
      SCHEDULE_REJECT = Symbol('schedule_reject'),
      SCHEDULE_DONE = Symbol('schedule_done'),
      SCHEDULE_SOON = Symbol('schedule_soon'),
      SCHEDULE_TICK = Symbol('schedule_tick');

/**
 * The AbstractPledge Class:
 * The base class for all custom Pledge classes
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 */
const AbstractPledge = Fn.inherits('Informer', function AbstractPledge(executor) {});

AbstractPledge.setStatic('RESOLVED_VALUE', RESOLVED_VALUE);

AbstractPledge.setProperty({

	/**
	 * The initial state is 0 (pending)
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.4.0
	 * @version  0.4.0
	 *
	 * @type     {Number}
	 */
	[STATE]: PENDING,

	/**
	 * Do not report progress by default, it's very expensive
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.7.0
	 * @version  0.7.0
	 *
	 * @type     {Boolean}
	 */
	report_progress: false,

	/**
	 * The pledge count
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.4.0
	 * @version  0.4.0
	 *
	 * @type     {Number}
	 */
	[SUB_PLEDGES]: 0,

	/**
	 * An array with pledges waiting for this pledge's fulfillment go here
	 * The initial _on_fulfilled property is null
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.4.0
	 * @version  0.4.0
	 */
	[ON_FULFILLED]: null,

	/**
	 * An array with pledges waiting for this pledge's rejection go here
	 * The initial _on_rejected property is null
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.4.0
	 * @version  0.4.0
	 */
	[ON_REJECTED]: null,

	/**
	 * The eventual resolved value
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.7.0
	 * @version  0.7.0
	 */
	[RESOLVED_VALUE]: null,

	/**
	 * When this pledge has ended
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.7.0
	 * @version  0.7.0
	 */
	[ENDED]: null,

	/**
	 * Last update to this pledge
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.7.1
	 * @version  0.7.1
	 *
	 * @type     {Number}
	 */
	[UPDATED]: null,

	/**
	 * Debug durations
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.7.1
	 * @version  0.7.1
	 *
	 * @type     {Object}
	 */
	_durations: null,

	/**
	 * Property that could be a function to cancel the pledge
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.7.0
	 * @version  0.7.0
	 *
	 * @type     {Function}
	 */
	cancel: null,

	/**
	 * Warn when an error is not caught
	 *
	 * @author   Jelle De Loecker   <jelle@elevenways.be>
	 * @since    0.8.1
	 * @version  0.8.1
	 *
	 * @type     {Boolean}
	 */
	warn_uncaught_errors: true,

	/**
	 * The 'nextTick' scheduler
	 *
	 * @author   Jelle De Loecker   <jelle@elevenways.be>
	 * @since    0.8.15
	 * @version  0.8.15
	 *
	 * @type     {Function}
	 */
	[SCHEDULE_TICK] : {value: Blast.nextTick},

	/**
	 * The 'setImmediate' scheduler
	 *
	 * @author   Jelle De Loecker   <jelle@elevenways.be>
	 * @since    0.8.15
	 * @version  0.8.15
	 *
	 * @type     {Function}
	 */
	[SCHEDULE_SOON] : {value: Blast.setImmediate},
});

/**
 * Return the state of the pledge
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @type     {Number}
 */
AbstractPledge.setProperty(function state() {
	return this[STATE];
});

/**
 * Has this pledge finished?
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.7.24
 * @version  0.7.24
 *
 * @type     {Boolean}
 */
AbstractPledge.setProperty(function is_done() {
	return this[STATE] > 0;
});

/**
 * Has this pledge ended?
 *
 * @deprecated
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @type     {Boolean}
 */
AbstractPledge.setProperty(function _ended() {
	return this.is_done;
});

/**
 * Determine if an object is an instance of Pledge
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.4.0
 *
 * @return   {Boolean}
 */
AbstractPledge.setStatic(function isPledge(obj) {
	return !!obj && typeof obj[SUB_PLEDGES] == 'number';
});

/**
 * Determine if an object is a thennable
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.7.0
 * @version  0.7.0
 *
 * @return   {Boolean}
 */
AbstractPledge.setStatic(function isThenable(obj) {
	return !!obj && typeof obj.then == 'function';
});

/**
 * Determine if an object has the promise interface
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.7.0
 * @version  0.7.0
 *
 * @return   {Boolean}
 */
AbstractPledge.setStatic(function hasPromiseInterface(obj) {
	return Pledge.isThenable(obj) && typeof obj.catch == 'function';
});

/**
 * Handle an old-style callback for a pledge or promise
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.7.0
 * @version  0.7.0
 */
AbstractPledge.setStatic(function done(promise, callback) {
	var result = Pledge.prototype.done.call(promise, callback);

	if (!result) {
		result = Pledge.resolve();
	}

	return result;
});

/**
 * Cast the given value to a pledge
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.7.0
 * @version  0.7.0
 *
 * @param    {Mixed}   value
 *
 * @return   {Pledge}
 */
AbstractPledge.setStatic(function cast(value) {

	var is_primitive = !value || Obj.isPrimitive(value);

	if (!is_primitive) {
		if (Pledge.isPledge(value)) {
			return value;
		}

		if (Pledge.hasPromiseInterface(value)) {
			let pledge = new Pledge();
			Pledge.done(value, pledge);
			return pledge;
		}
	}

	return Pledge.resolve(value);

});

/**
 * Create a new pledge and resolve it with the given value
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.6
 * @version  0.8.15
 *
 * @param    {Mixed}   value
 *
 * @return   {Pledge}
 */
AbstractPledge.setStatic(function resolve(value) {

	const pledge = new this();

	pledge[SCHEDULE_TICK](function onNextTick() {
		pledge.resolve(value);
	});

	return pledge;
});

/**
 * Create a new pledge and reject it with the given value
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.6
 * @version  0.5.6
 *
 * @param    {Error}   err
 *
 * @return   {Pledge}
 */
AbstractPledge.setStatic(function reject(err) {

	const pledge = new this();

	pledge[SCHEDULE_TICK](function onNextTick() {
		pledge.reject(err);
	});

	return pledge;
});

/**
 * Create a new pledge and perform the given tasks
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.6
 * @version  0.5.6
 *
 * @param    {Array}   tasks
 *
 * @return   {Pledge}
 */
AbstractPledge.setStatic(function all(tasks) {

	if (tasks == null || !Array.isArray(tasks)) {
		return this.reject(new Error('No valid tasks were given'));
	}

	return Fn.parallel(false, tasks);
});

/**
 * Create a new pledge and race the given tasks
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.6
 * @version  0.5.6
 *
 * @param    {Array}   tasks
 *
 * @return   {Pledge}
 */
AbstractPledge.setStatic(function race(tasks) {
	return new this(function doPledgeRace(resolve, reject) {
		for (var i = 0; i < tasks.length; i += 1) {
			Pledge.resolve(tasks[i]).then(resolve, reject);
		}
	});
});

/**
 * Create a pledge which resolves after n milliseconds
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.6.0
 * @version  0.6.0
 *
 * @param    {Number}   n
 * @param    {Mixed}    value
 *
 * @return   {Pledge}
 */
AbstractPledge.setStatic(function after(n, value) {

	var pledge = new Pledge(),
	    id;

	id = setTimeout(function doResolve() {
		pledge.resolve(value);
	}, n);

	return pledge;
});

/**
 * Schedule something to do with the resolver
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Function}    fnc
 */
AbstractPledge.setMethod(SCHEDULE_RESOLVE, function(fnc) {
	return this[SCHEDULE_TICK](fnc);
});

/**
 * Schedule something to do with the rejector
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Function}    fnc
 */
AbstractPledge.setMethod(SCHEDULE_REJECT, function(fnc) {
	return this[SCHEDULE_SOON](fnc);
});

/**
 * Schedule something to do with the executor
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Function}    fnc
 */
AbstractPledge.setMethod(SCHEDULE_EXECUTOR, function(fnc) {
	return this[SCHEDULE_SOON](fnc);
});

/**
 * Schedule something to do with handling the result
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Function}    fnc
 */
AbstractPledge.setMethod(SCHEDULE_DONE, function(...args) {
	return this[SCHEDULE_TICK](...args);
});

/**
 * Add progress amount (dummy)
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Number}   value   A value between 0-100
 * @param    {String}   label   An optional label
 */
AbstractPledge.setMethod(function addProgress(value, label) {});

/**
 * Add an amount of progress parts used for calculating the progress (dummy)
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Number}   parts
 */
AbstractPledge.setMethod(function addProgressPart(parts) {});

/**
 * Finish an amount of progress parts (dummy)
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Number}   parts
 */
AbstractPledge.setMethod(function reportProgressPart(parts) {});

/**
 * Add progress pledge (dummy)
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Pledge}   pledge
 */
AbstractPledge.setMethod(function _addProgressPledge(pledge) {});

/**
 * The BasePledge Class
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.6.4
 *
 * @param    {Function}   executor
 */
const Pledge = Fn.inherits(AbstractPledge, function Pledge(executor) {

	this[CREATED] = Date.now();

	if (!executor) {
		return;
	}

	// Store the executor
	this[EXECUTOR] = executor;

	// Start the executor when the event loop is empty
	this[START_EXECUTOR](false);
});

Pledge.setProperty({

	/**
	 * When the pledge was created
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.7.0
	 * @version  0.7.0
	 *
	 * @type     {Number}
	 */
	[CREATED]: 0,

	/**
	 * The initial progress is 0
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.4.0
	 * @version  0.4.0
	 *
	 * @type     {Number}
	 */
	progress: 0,

	/**
	 * Progress parts can be used instead of manually adding percentages
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.5.2
	 * @version  0.5.2
	 *
	 * @type     {Number}
	 */
	progress_parts: 0,

	/**
	 * The finished number of progress parts
	 *
	 * @author   Jelle De Loecker   <jelle@develry.be>
	 * @since    0.5.2
	 * @version  0.5.2
	 *
	 * @type     {Number}
	 */
	progress_parts_finished: 0,

});

/**
 * Calculate the duration
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.6.4
 * @version  0.6.4
 *
 * @type     {Number}
 */
Pledge.setProperty(function duration() {

	var start = this[STARTED] || this[CREATED],
	    end = this[ENDED] || Date.now();

	return end - start;
});

/**
 * Calculated end time
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.6.4
 * @version  0.6.4
 *
 * @type     {Date}
 */
Pledge.setProperty(function due_date() {

	if (!this.progress_parts && !this.progress) {
		return null;
	}

	let duration = this.duration;

	if (duration > 0) {

		let expected_duration;

		if (this.progress_parts) {
			expected_duration = (duration / (this.progress_parts_finished || 1)) * this.progress_parts;
		} else if (this.progress) {
			expected_duration = duration * (100 / this.progress);
		}

		return new Date((this[STARTED] || this[CREATED]) + expected_duration);
	}

	return null;
});

/**
 * Start the executor
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.6.4
 *
 * @param    {Boolean}   do_async   Do the executor asynchronously
 */
Pledge.setMethod(START_EXECUTOR, function _startExecutor(do_async) {

	// Don't start the executor twice!
	if (this[STARTED]) {
		return;
	}

	if (do_async === true) {
		return this[SCHEDULE_EXECUTOR](() => this[START_EXECUTOR]());
	}

	// Indicate the resolver has started
	this[STARTED] = Date.now();

	// Don't do anything if there is no executor
	if (!this[EXECUTOR]) {
		return;
	}

	// Do the resolver in the current pledge context
	this[EXECUTOR](value => this.resolve(value), reason => this.reject(reason));
});

/**
 * Add progress amount
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.7.0
 *
 * @param    {Number}   value   A value between 0-100
 * @param    {String}   label   An optional label
 */
Pledge.setMethod(function addProgress(value, label) {

	if (!this.report_progress) {
		return;
	}

	return this.reportProgress((this.progress || 0) + value, label);
});

/**
 * Report progress
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.4.0
 *
 * @param    {Number}   value   A value between 0-100
 * @param    {String}   label   An optional label
 */
Pledge.setMethod(function reportProgress(value, label) {
	this.progress = value;
	return this.progress;
});

/**
 * Add an amount of progress parts used for calculating the progress
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.2
 * @version  0.5.2
 *
 * @param    {Number}   parts
 */
Pledge.setMethod(function addProgressPart(parts) {

	if (!this.report_progress) {
		return;
	}

	if (parts == null) {
		parts = 1;
	}

	this.progress_parts += parts;
	this.calculateProgressParts();
});

/**
 * Finish an amount of progress parts
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.2
 * @version  0.7.1
 *
 * @param    {Number}   parts
 */
Pledge.setMethod(function reportProgressPart(parts) {

	if (!this.report_progress) {
		return;
	}

	if (parts == null) {
		parts = 1;
	}

	// Get the current timestamp
	let now = Date.now();

	// Get the previous update timestamp
	let last_update = this[UPDATED] || this[CREATED] || now;

	// Set the new updated time
	this[UPDATED] = now;

	if (!this._durations) {
		this._durations = {};
	}

	this._durations[this.progress_parts_finished] = now - last_update;

	this.progress_parts_finished += parts;
	this.calculateProgressParts();
});

/**
 * Calculate the progress based on the progress parts
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.2
 * @version  0.6.3
 */
Pledge.setMethod(function calculateProgressParts() {

	var amount;

	if (this[STATE] == PENDING) {
		// Calculate how much is already finished
		amount = this.progress_parts_finished / this.progress_parts;

		// And multiply it to get a nice percentage number
		amount = ~~(amount * 100);
	} else {
		amount = 100;
	}

	this.reportProgress(amount);
});

/**
 * Add progress pledge
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.6.3
 * @version  0.6.3
 *
 * @param    {Pledge}   pledge
 */
Pledge.setMethod(function _addProgressPledge(pledge) {

	if (!this.report_progress) {
		return;
	}

	if (!Pledge.isPledge(pledge)) {
		return false;
	}

	this.addProgressPart(pledge.progress_parts);
	this.reportProgressPart(pledge.progress_parts_finished);

	let that = this,
	    other_addProgressPart = pledge.addProgressPart,
	    other_reportProgressPart = pledge.reportProgressPart;

	pledge.addProgressPart = function addProgressPart(parts) {
		that.addProgressPart(parts);
		other_addProgressPart.call(this, parts);
	};

	pledge.reportProgressPart = function reportProgressPart(parts) {
		that.reportProgressPart(parts);
		other_reportProgressPart.call(this, parts);
	};
});

/**
 * Resolve with the given value
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.5.6
 */
Pledge.setMethod(function resolve(value) {

	if (this[STATE] != PENDING) {
		return;
	}

	if (value === this) {
		throw new TypeError('A pledge cannot be resolved with itself.');
	}

	this[DO_RESOLVE](value);
});

/**
 * Do the actual resolving
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.6.4
 */
Pledge.setMethod(DO_RESOLVE, function _doResolve(value) {

	var that = this;

	if (value && value.then) {

		this._addProgressPledge(value);

		return value.then(function onFulfilledResolve(value) {
			that.resolve(value);
		}, function onRejectedResolve(reason) {
			that.reject(reason);
		});
	}

	this[STATE] = RESOLVED;
	this[RESOLVED_VALUE] = value;
	this[ENDED] = Date.now();

	this[SCHEDULE_RESOLVE](function resolveOnNextTick() {
		if (that[SUB_PLEDGES] !== 0) {
			while (that[ON_FULFILLED].length) {
				that[ON_FULFILLED].shift()(value);
			}

			that[ON_REJECTED].length = 0;
		}
	});

	if (this.report_progress) {
		this.calculateProgressParts();
	}
});

/**
 * Reject with the given reason
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.8.1
 *
 * @param    {Error}   reason
 */
Pledge.setMethod(function reject(reason) {

	if (this[STATE] != PENDING) {
		return;
	}

	const that = this;

	this[SCHEDULE_REJECT](function doRejection() {
		let result = that[DO_REJECT](reason, false);

		if (!result && that.warn_uncaught_errors) {
			console.warn('Uncaught Pledge error:', reason);
		}
	});
});

/**
 * Do the actual rejection
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.6.4
 *
 * @param    {Error}   reason
 * @param    {Boolean} do_async
 */
Pledge.setMethod(DO_REJECT, function _doReject(reason, do_async) {

	var that = this,
	    caught_rejection = false;

	if (reason && reason.then) {
		return reason.then(function onFulfilledReject(value) {
			that.reject(value);
		}, function onRejectedReject(reason) {
			that.reject(reason);
		});
	}

	this[STATE] = REJECTED;
	this[REJECTED_REASON] = reason;
	this[ENDED] = Date.now();

	function resolveOnNextTick() {

		if (that[SUB_PLEDGES] !== 0) {
			while (that[ON_REJECTED].length) {
				if (that[ON_REJECTED].shift()(reason)) {
					caught_rejection = true;
				}
			}

			that[ON_FULFILLED].length = 0;
		}
	}

	if (do_async) {
		this[SCHEDULE_TICK](resolveOnNextTick);
		return true;
	}

	resolveOnNextTick();

	return caught_rejection;
});

/**
 * Then
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.7.1
 *
 * @return   {Pledge}
 */
Pledge.setMethod(function then(on_fulfilled, on_rejected) {

	var that = this,
	    then_pledge = new Pledge();

	// Create the arrays only when listeners are added
	if (this[SUB_PLEDGES] == 0) {
		this[ON_FULFILLED] = [];
		this[ON_REJECTED] = [];
	}

	this[SUB_PLEDGES]++;

	this[ON_FULFILLED].push(function onFulfilled(value) {

		var result;

		try {
			if (on_fulfilled) {
				result = on_fulfilled(value);
			} else {
				result = value;
			}

			then_pledge.resolve(result);
		} catch (reason) {
			then_pledge.reject(reason);
		}
	});

	this[ON_REJECTED].push(function onRejected(reason) {

		var caught_rejection = false,
		    result;

		try {
			if (on_rejected) {
				result = on_rejected(reason);
				caught_rejection = true;
				that._caught_rejection = true;
			} else {
				result = reason;
			}

			// This was wrong in earlier versions:
			// you RESOLVE chained promises, not reject
			then_pledge.resolve();
		} catch (reason) {
			then_pledge.reject(reason);
		}

		return caught_rejection;
	});

	if (this[STATE] == RESOLVED) {
		this[DO_RESOLVE](this[RESOLVED_VALUE]);
	} else if (this[STATE] == REJECTED) {
		this[DO_REJECT](this[REJECTED_REASON]);
	}

	return then_pledge;
});

/**
 * Catch
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.4.0
 * @version  0.4.0
 *
 * @return   {Pledge}
 */
Pledge.setMethod('catch', function _catch(on_rejected) {
	return this.then(null, on_rejected);
});

/**
 * When the promise is settled, whether fulfilled or rejected,
 * the specified callback function is executed
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.6
 * @version  0.5.6
 *
 * @param    {Function}   on_finally
 *
 * @return   {Pledge}
 */
Pledge.setMethod('finally', function _finally(on_finally) {

	var constructor = this.constructor;

	return this.then(
		function afterResolved(value) {
			return constructor.resolve(on_finally()).then(function() {
				return value;
			});
		},
		function afterRejected(reason) {
			return constructor.resolve(on_finally()).then(function() {
				return constructor.reject(reason);
			});
		}
	);
});

/**
 * Handle an old-style callback
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.5.0
 * @version  0.7.0
 *
 * @return   {Pledge}
 */
Pledge.setMethod(['done', 'handleCallback'], function done(callback) {

	if (!callback) {
		return this;
	}

	if (typeof callback != 'function') {
		if (Pledge.isPledge(callback)) {
			let pledge = callback;

			callback = function forwardToPledge(err, result) {
				if (err) {
					pledge.reject(err);
				} else {
					pledge.resolve(result);
				}
			}
		} else {
			throw new Error('Unable to handle callback: not a function or Pledge!');
		}
	}

	// Catch when the supplied context is changed to a non-promise
	if (!this || !this.then) {
		return Blast.nextTick(callback, null, null, this);
	}

	return this.then(function onResolved(value) {
		return callback(null, value);
	}, function onRejected(err) {
		return callback(err);
	});
});

/**
 * Race this pledge with another one
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.6.0
 * @version  0.6.0
 *
 * @param    {Pledge}   contestant
 */
Pledge.setMethod(function race(contestant) {
	var contest = Pledge.race([this].concat(contestant));
	return contest;
});

/**
 * The LazyPledge Class
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.6.3
 * @version  0.6.3
 *
 * @param    {Function}   executor
 */
var Lazy = Collection.Function.inherits('Pledge', function LazyPledge(executor) {
	this[EXECUTOR] = executor;
});

/**
 * Then
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.6.3
 * @version  0.6.3
 *
 * @return   {Pledge}
 */
Lazy.setMethod(function then(on_fulfilled, on_rejected) {

	var result = then.super.call(this, on_fulfilled, on_rejected);

	if (typeof this[EXECUTOR] != 'function') {
		this.reject(new Error('No valid executor has been given'));
	} else {
		this[START_EXECUTOR]();
	}

	return result;
});

/**
 * The Timeout Pledge Class
 *
 * @author   Jelle De Loecker   <jelle@develry.be>
 * @since    0.7.1
 * @version  0.7.10
 *
 * @param    {Function}   executor
 * @param    {Number}     timeout
 * @param    {String}     message
 */
var Timeout = Collection.Function.inherits('Pledge', function TimeoutPledge(executor, timeout, message) {

	const that = this;

	if (typeof executor == 'number') {
		message = timeout;
		timeout = executor;
		executor = null;
	}

	if (typeof timeout != 'number') {
		timeout = +timeout;
	}

	if (!timeout) {
		timeout = 0;
	}

	setTimeout(function checkTimeout() {

		if (that[STATE] != PENDING) {
			return;
		}

		if (!message) {
			message = 'Timeout of ' + timeout + 'ms was reached';
		}

		let error = new Error(message);

		that.reject(error);

	}, timeout);

	TimeoutPledge.super.call(this, executor);
});

/**
 * The SwiftPledge Class
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Function}   executor
 */
const Swift = Collection.Function.inherits(Pledge, function SwiftPledge(executor) {
	this[EXECUTOR] = executor;
	this[START_EXECUTOR](false);
});

const SwiftObject = Object.create(Bound.Function);
SwiftObject[Blast.flowPledgeClass] = Swift;
SwiftObject[Blast.asyncScheduler] = Blast.callNow;

/**
 * Execute the given function and return the value as swiftly as possible.
 * If no promise is returned, the value will be returned as is.
 * If the given value is not a function, it will be returned as is.
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Mixed}   value
 *
 * @return   {SwiftPledge|Mixed}
 */
Swift.setStatic(function execute(value) {

	if (typeof value == 'function') {
		value = value();
	}

	if (value && value.then) {

		if (value[STATE] == RESOLVED) {
			return execute(value[RESOLVED_VALUE]);
		}

		return Swift.resolve(value);
	}

	return value;
});

/**
 * Perform a swift waterfall
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @return   {SwiftPledge|Mixed}
 */
Swift.setStatic(function waterfall(...tasks) {
	let result = SwiftObject.waterfall(...tasks);
	return Swift.execute(result);
});

/**
 * Create a new SwiftPledge and resolve it with the given value
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Mixed}   value
 *
 * @return   {Pledge}
 */
Swift.setStatic(function resolve(value) {
	let result = new this();
	result.resolve(value);
	return result;
});

/**
 * Start the executor
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Boolean}   do_async   Do the executor asynchronously
 */
Swift.setMethod(START_EXECUTOR, function _startExecutor(do_async) {

	// Don't start the executor twice!
	if (this[STARTED]) {
		return;
	}

	// Don't do anything if there is no executor
	if (!this[EXECUTOR]) {
		return;
	}

	this[STARTED] = true;

	if (do_async === true) {
		return Blast.nextGroupedImmediate(() => this[START_EXECUTOR]());
	}

	// Do the resolver in the current pledge context
	this[EXECUTOR](value => this.resolve(value), reason => this.reject(reason));
});

/**
 * Do the actual resolving
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 */
Swift.setMethod(DO_RESOLVE, function _doResolve(value) {

	if (value && value.then) {
		return value.then(value => this.resolve(value), reason => this.reject(reason));
	}

	this[STATE] = RESOLVED;
	this[RESOLVED_VALUE] = value;

	if (this[SUB_PLEDGES] !== 0) {
		while (this[ON_FULFILLED].length) {
			this[ON_FULFILLED].shift()(value);
		}

		this[ON_REJECTED].length = 0;
	}
});


/**
 * Reject with the given reason
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Error}   reason
 */
Swift.setMethod(function reject(reason) {

	if (this[STATE] != PENDING) {
		return;
	}

	this[DO_REJECT](reason);
});

/**
 * Do the actual rejection
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @param    {Error}   reason
 */
Swift.setMethod(DO_REJECT, function _doReject(reason) {

	let caught_rejection = false;

	if (reason && reason.then) {
		return reason.then(value => this.reject(value), reason => this.reject(reason));
	}

	this[STATE] = REJECTED;
	this[REJECTED_REASON] = reason;

	if (this[SUB_PLEDGES] !== 0) {
		while (this[ON_REJECTED].length) {
			if (this[ON_REJECTED].shift()(reason)) {
				caught_rejection = true;
			}
		}

		this[ON_FULFILLED].length = 0;
	}

	return caught_rejection;
});

/**
 * Then
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 *
 * @return   {Pledge}
 */
Swift.setMethod(function then(on_fulfilled, on_rejected) {

	// Create the arrays only when listeners are added
	if (this[SUB_PLEDGES] == 0) {
		this[ON_FULFILLED] = [];
		this[ON_REJECTED] = [];
	}

	this[SUB_PLEDGES]++;

	if (on_fulfilled) {
		this[ON_FULFILLED].push(on_fulfilled);
	}

	if (on_rejected) {
		this[ON_REJECTED].push(on_rejected);
	}

	if (this[STATE] == RESOLVED) {
		this[DO_RESOLVE](this[RESOLVED_VALUE]);
	} else if (this[STATE] == REJECTED) {
		this[DO_REJECT](this[REJECTED_REASON]);
	}

	return this;
});

/**
 * Create an already resolved pledge
 *
 * @author   Jelle De Loecker   <jelle@elevenways.be>
 * @since    0.8.15
 * @version  0.8.15
 */
function createResolved(PledgeClass) {
	PledgeClass.setStatic('RESOLVED', PledgeClass.resolve(true), false);
}

Blast.defineClass('Pledge', Pledge);
Blast.defineClass('LazyPledge', Lazy);
Blast.defineClass('SwiftPledge', Swift);

createResolved(Pledge);
createResolved(Swift);

Pledge.Lazy = Lazy;
Pledge.Timeout = Timeout;
Pledge.Swift = Swift;